from roboticstoolbox.mobile.vehicle import *
from roboticstoolbox.mobile.bug2 import Bug2
from roboticstoolbox.mobile.dx_form import DXform
from roboticstoolbox.mobile.dstar import DStar
from roboticstoolbox.mobile.landmarkmap import LandmarkMap
from roboticstoolbox.mobile.sensors import RangeBearingSensor
from roboticstoolbox.mobile.drivers import *
from roboticstoolbox.mobile.animations import *
from roboticstoolbox.mobile.PoseGraph import *

__all__ = [
    'Vehicle',
    'Bicycle',
    'Unicycle',
    'Bug2',
    'DXform',
    'DStar',
    'VehicleMarker',
    'VehiclePolygon',
    'VehicleIcon',
    'VehicleDriver',
    'RandomPath',
    'PurePursuit',
    'LandmarkMap',
    'RangeBearingSensor',
    'PoseGraph',
]

# __doc__ = """
# The Robotics Toolbox for Python
# Based on the Matlab version
# Peter Corke 2007
# """

######################
#   Import Section   #
######################

# from numpy import *

# # Import Link Constructor section
# from Link import *

# # Import Robot Constructor section
# from Robot import *

# # utility
# from utility import *

# # Import transformations section
# from transform import *

# from trplot import *

# from jacobian import *

# # import kinematics section
# from kinematics import *

# from manipulability import *

# # import trajectories section
# from trajectory import *

# # import Quaternion constructor section
# from Quaternion import *

# # import dynamics section
# from dynamics import *

# # import robot models sections
# #from puma560 import *
# #from puma560akb import *
# #from stanford import *
# #from twolink import *

# print """
# Robotics Toolbox for Python
# Based on Matlab Toolbox Version 7  April-2002

# What's new.
#   Readme      - New features and enhancements in this version.

# Homogeneous transformations
#   eul2tr      - Euler angle to transform
#   oa2tr       - orientation and approach vector to transform
#   rotx        - transform for rotation about X-axis
#   roty        - transform for rotation about Y-axis
#   rotz        - transform for rotation about Z-axis
#   rpy2tr      - roll/pitch/yaw angles to transform
#   tr2eul      - transform to Euler angles
#   tr2rot      - transform to rotation submatrix
#   tr2rpy      - transform to roll/pitch/yaw angles
#   transl      - set or extract the translational component of a transform
#   trnorm      - normalize a transform

# Quaternions
#   /           - divide quaternion by quaternion or scalar
#   *           - multiply quaternion by a quaternion or vector
#   inv         - invert a quaternion
#   norm        - norm of a quaternion
#   plot        - display a quaternion as a 3D rotation
#   qinterp     - interpolate quaternions
#   unit        - unitize a quaternion

# Kinematics
#   diff2tr     - differential motion vector to transform
#   fkine       - compute forward kinematics
#   ikine       - compute inverse kinematics
#   ikine560    - compute inverse kinematics for Puma 560 like arm
#   jacob0      - compute Jacobian in base coordinate frame
#   jacobn      - compute Jacobian in end-effector coordinate frame
#   tr2diff     - transform to differential motion vector
#   tr2jac      - transform to Jacobian

# Dynamics
#   accel       - compute forward dynamics
#   cinertia    - compute Cartesian manipulator inertia matrix
#   coriolis    - compute centripetal/coriolis torque
#   friction    - joint friction
#   ftrans      - transform force/moment
#   gravload    - compute gravity loading
#   inertia     - compute manipulator inertia matrix
#   itorque     - compute inertia torque
#   nofriction  - remove friction from a robot object
#   rne         - inverse dynamics

# Trajectory generation
#   ctraj       - Cartesian trajectory
#   jtraj       - joint space trajectory
#   trinterp    - interpolate transform s

# Graphics
#   drivebot    - drive a graphical  robot
#   plot        - plot/animate robot

# Other
#   manipblty   - compute manipulability
#   unit        - unitize a vector

# Creation of robot models.
#   link        - construct a robot link object
#   puma560     - Puma 560 data
#   puma560akb  - Puma 560 data (modified Denavit-Hartenberg)
#   robot       - construct a robot object
#   stanford    - Stanford arm data
#   twolink     - simple 2-link example
# """



from roboticstoolbox import ETS as ET
from roboticstoolbox import ERobot


class Puma560(ERobot):
    """
    Create model of Franka-Emika Panda manipulator

    ``puma = Puma560()`` creates a robot object representing the classic
    Unimation Puma560 robot arm. This robot is represented using the elementary
    transform sequence (ETS).

    .. note::

        - The model has different joint offset conventions compared to
          ``DH.Puma560()``. For this robot:
            - Zero joint angles ``qz`` is the vertical configuration,
              corresponding to ``qr`` with ``DH.Puma560()``
            - ``qbent`` is the bent configuration, corresponding to
              ``qz`` with ``DH.Puma560()``

    :references:
        - "A Simple and Systematic Approach to Assigning Denavit–Hartenberg
          Parameters,", P. I. Corke,  in IEEE Transactions on Robotics, vol. 23,
          no. 3, pp. 590-594, June 2007, doi: 10.1109/TRO.2007.896765.
        - https://petercorke.com/robotics/a-simple-and-systematic-approach-to-assigning-denavit-hartenberg-parameters

    """  # noqa

    def __init__(self):
        # Puma dimensions (m)
        l1 = 0.672
        l2 = 0.2337
        l3 = 0.4318
        l4 = -0.0837
        l5 = 0.4318
        l6 = 0.0203

        e = ET.tz(l1) * ET.rz() * ET.ty(l2) * ET.ry() * ET.tz(l3) * \
            ET.tx(l6) * ET.ty(l4) * ET.ry() * ET.tz(l5) * ET.rz() * \
            ET.ry() * ET.rz() * ET.tx(0.2)

        super().__init__(
            e,
            name='Puma560',
            manufacturer='Unimation',
            comment='ETS-based model')

        self.addconfiguration(
            "qz", [0, 0, 0, 0, 0, 0])
        self.addconfiguration(
            "qbent", [0, -90, 90, 0, 0, 0], 'deg')


if __name__ == '__main__':   # pragma nocover

    robot = Puma560()
    print(robot)



#!/usr/bin/env python

import numpy as np
from roboticstoolbox.robot.ETS import ETS
from roboticstoolbox.robot.ERobot import ERobot
from roboticstoolbox.robot.ELink import ELink


class GenericSeven(ERobot):
    """
    Create model of a generic seven degree-of-freedom robot

    robot = GenericSeven() creates a robot object. This robot is represented
    using the elementary transform sequence (ETS).

    """
    def __init__(self):

        deg = np.pi/180
        mm = 1e-3
        tool_offset = (103)*mm

        link_length = 0.5

        l0 = ELink(
            ETS.tz(link_length) * ETS.rz(),
            name='link0',
            parent=None
        )

        l1 = ELink(
            ETS.ry(),
            name='link1',
            parent=l0
        )

        l2 = ELink(
            ETS.tz(link_length) * ETS.rz(),
            name='link2',
            parent=l1
        )

        l3 = ELink(
            ETS.ry(),
            name='link3',
            parent=l2
        )

        l4 = ELink(
            ETS.tz(link_length) * ETS.rz(),
            name='link4',
            parent=l3
        )

        l5 = ELink(
            ETS.ry(),
            name='link5',
            parent=l4
        )

        l6 = ELink(
            ETS.tx(link_length) * ETS.rz(),
            name='link6',
            parent=l5
        )

        ee = ELink(
            ETS.tz(-link_length),
            name='ee',
            parent=l6
        )

        elinks = [l0, l1, l2, l3, l4, l5, l6, ee]
        # elinks = [l0, l1, l2, l3, l4, l5, ee]

        super(GenericSeven, self).__init__(
            elinks,
            name='Generic Seven',
            manufacturer="Jesse's Imagination")

        # self.addconfiguration(
        #     "qz", np.array([0, 0, 0, 0, 0, 0, 0]))
        # self.addconfiguration(
        #     "qr", np.array([0, -0.3, 0, -2.2, 0, 2.0, np.pi/4]))



#!/usr/bin/env python

import numpy as np
from roboticstoolbox.robot.ERobot import ERobot


class UR10(ERobot):
    """
    Class that imports a UR10 URDF model

    ``UR3()`` is a class which imports a Universal Robotics UR310 robot
    definition from a URDF file.  The model describes its kinematic and
    graphical characteristics.

    .. runblock:: pycon

        >>> import roboticstoolbox as rtb
        >>> robot = rtb.models.URDF.UR10()
        >>> print(robot)

    Defined joint configurations are:

    - qz, zero joint angle configuration, 'L' shaped configuration
    - qr, vertical 'READY' configuration

    .. codeauthor:: Jesse Haviland
    .. sectionauthor:: Peter Corke
    """

    def __init__(self):

        args = super().urdf_to_ets_args(
            "ur_description/urdf/ur10_joint_limited_robot.urdf.xacro"
        )

        super().__init__(
                args[0],
                name=args[1],
                manufacturer='Universal Robotics'
            )

        self.addconfiguration(
            "qz", np.array([0, 0, 0, 0, 0, 0]))
        self.addconfiguration(
            "qr", np.array([np.pi, 0, 0, 0, np.pi/2, 0]))


if __name__ == '__main__':   # pragma nocover

    robot = UR10()
    print(robot)



#!/usr/bin/env python
"""
@author Jesse Haviland
"""

import numpy as np
from spatialmath.base.argcheck import getvector


class EllipsePlot(object):

    def __init__(self, robot, q, etype, opt='trans', centre=[0, 0, 0], scale=0.1):

        super(EllipsePlot, self).__init__()

        try:
            centre = getvector(centre, 3)
        except ValueError:
            centre = getvector(centre, 2)
            centre = np.array([centre[0], centre[1], 0])
        except TypeError:
            if centre != 'ee':
                raise ValueError(
                    'Centre must be a three vector or \'ee\' meaning'
                    'end-effector')

        self.ell = None
        self.robot = robot
        self.opt = opt
        self.centre = centre
        self.ax = None
        self.scale = scale

        if q is None:
            self.q = robot.q
        else:
            self.q = q

        if etype == 'v':
            self.vell = True
            self.name = 'Velocity Ellipse'
        elif etype == 'f':
            self.vell = False
            self.name = 'Force Ellipse'

    def draw(self):
        self.make_ellipsoid()

        if self.ell is not None:
            self.ax.collections.remove(self.ell)

        self.ell = self.ax.plot_wireframe(
            self.x, self.y, self.z, rstride=6,
            cstride=6, color='#2980b9', alpha=0.2)

    def draw2(self):
        self.make_ellipsoid2()

        if self.ell is not None:
            self.ell[0].set_data(self.x, self.y)
        else:
            self.ell = self.ax.plot(
                self.x, self.y, color='#2980b9', alpha=0.2)

    def make_ellipsoid(self):
        """
        Plot the 3d Ellipsoid

        """

        if self.opt == 'trans':
            J = self.robot.jacobe(self.q)[3:, :]
            A = J @ J.T
        elif self.opt == 'rot':
            J = self.robot.jacobe(self.q)[:3, :]
            A = J @ J.T

        if not self.vell:
            # Do the extra step for the force ellipse
            A = np.linalg.inv(A)

        if isinstance(self.centre, str) and self.centre == 'ee':
            centre = self.robot.fkine(self.q).t
        else:
            centre = self.centre

        # find the rotation matrix and radii of the axes
        U, s, rotation = np.linalg.svd(A)
        radii = 1.0 / np.sqrt(s)

        # points on unit sphere
        u = np.linspace(0.0, 2.0 * np.pi, 50)
        v = np.linspace(0.0, np.pi, 50)
        x = radii[0] * np.outer(np.cos(u), np.sin(v))
        y = radii[1] * np.outer(np.sin(u), np.sin(v))
        z = radii[2] * np.outer(np.ones_like(u), np.cos(v))

        # transform points to ellipsoid
        for i in range(len(x)):
            for j in range(len(x)):
                [x[i, j], y[i, j], z[i, j]] = \
                    np.dot([x[i, j], y[i, j], z[i, j]], rotation)

        self.x = x * self.scale + centre[0]
        self.y = y * self.scale + centre[1]
        self.z = z * self.scale + centre[2]

    def make_ellipsoid2(self):
        """
        Plot the 2d Ellipsoid

        """

        if self.opt == 'trans':
            J = self.robot.jacobe(self.q)[:2, :]
            A = J @ J.T
        elif self.opt == 'rot':
            raise ValueError(
                "Can not do rotational ellipse for a 2d robot plot."
                " Set opt='trans'")

        if not self.vell:
            # Do the extra step for the force ellipse
            A = np.linalg.inv(A)

        if isinstance(self.centre, str) and self.centre == 'ee':
            centre = self.robot.fkine(self.q).t
        else:
            centre = self.centre

        # find the rotation matrix and radii of the axes
        U, s, rotation = np.linalg.svd(A)
        radii = 1.0 / np.sqrt(s)

        # points on unit sphere
        u = np.linspace(0.0, 2.0 * np.pi, 50)
        v = np.linspace(0.0, np.pi, 50)
        x = radii[0] * np.outer(np.cos(u), np.sin(v))
        y = radii[1] * np.outer(np.sin(u), np.sin(v))

        # transform points to ellipsoid
        for i in range(len(x)):
            for j in range(len(x)):
                [x[i, j], y[i, j]] = \
                    np.dot([x[i, j], y[i, j]], rotation)

        self.x = x * self.scale + centre[0]
        self.y = y * self.scale + centre[1]



from abc import ABC
import numpy as np
import scipy as sp
from math import pi, sin, cos
import matplotlib.pyplot as plt
from spatialmath import base
import roboticstoolbox as rtb

class LandmarkMap:
    """
    LandmarkMap Map of planar point landmarks

    A LandmarkMap object represents a square 2D environment with a number of landmark
    landmark pointself.

    Methods::
      plot      Plot the landmark map
      landmark   Return a specified map landmark
      display   Display map parameters in human readable form
      char      Convert map parameters to human readable string

    Properties::
      map         Matrix of map landmark coordinates 2xN
      dim         The dimensions of the map region x,y in [-dim,dim]
      nlandmarks   The number of map landmarks N

    Examples::

    To create a map for an area where X and Y are in the range -10 to +10 metres
    and with 50 random landmark points
           map = LandmarkMap(50, 10)
    which can be displayed by
           self.plot()

    Reference::

      Robotics, Vision & Control, Chap 6,
      Peter Corke,
      Springer 2011

    See also RangeBearingSensor, EKF.
    """
    # TODO:
    # add a name property, show in char()

    #     properties
    #         map    # map landmarks
    #         dim         # map dimension
    #         nlandmarks   # number of landmarks in map

    #         verbose
    #     end


    def __init__(self, nlandmarks, dim=10, verbose=True, seed=None):
        """
        Create a map of point landmark landmarks
        %
        # M = LandmarkMap(N, DIM, OPTIONS) is a LandmarkMap object that represents N random point landmarks
        # in a planar region bounded by +/-DIM in the x- and y-directionself.
        %
        # Options::
        # 'verbose'    Be verbose
            
        
        %# TODO: dim can be a 4-vector
        """
    
        self._nlandmarks = nlandmarks
        self._dim = base.expand_dims(dim)

        random = np.random.default_rng(seed)
        x = random.uniform(self._dim[0], self._dim[1], nlandmarks)
        y = random.uniform(self._dim[2], self._dim[3], nlandmarks)

        self._map = np.c_[x, y].T
        self._verbose = verbose


    def __str__(self):
    # s = M.char() is a string showing map parameters in 
    # a compact human readable format. 
        s = f"LandmarkMap object with {self._nlandmarks} landmarks, dim=" + str(self._dim)
        return s

    @property
    def nlandmarks(self):
        return self._nlandmarks

    @property
    def x(self):
        return self._map[0,:]

    @property
    def y(self):
        return self._map[1,:]

    @property
    def xy(self):
        return self._map

    def landmark(self, k):
        """
        Get k'th landmarks from map
        %
        # F = M.landmark(K) is the coordinate (2x1) of the K'th landmark (landmark).
        """
        return self._map[:,k]

    def plot(self, block=False, **kwargs):
        """
        Plot the map
        %
        # M.plot() plots the landmark map in the current figure, as a square
        # region with dimensions given by the M.dim property.  Each landmark
        # is marked by a black diamond.
        %
        # M.plot(LS) as above, but the arguments LS
        # are passed to plot and override the default marker style.
        %
        # Notes::
        # - The plot is left with HOLD ON.
        """
        ax = base.plotvol2(self._dim)
        ax.set_aspect('equal')

        ax.set_xlabel('x')
        ax.set_ylabel('y')

        if len(kwargs) == 0:
            kwargs = {
                'linewidth': 0,
                'marker': 'P',
                'color': 'k',
            }

        plt.plot(self._map[0,:], self._map[1,:], **kwargs)
        plt.grid(True)
        plt.show(block=block)

if __name__ == "__main__":
    import unittest




# Copyright (c) 2015, Open Source Robotics Foundation, Inc.
# Copyright (c) 2013, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Open Source Robotics Foundation, Inc.
#       nor the names of its contributors may be used to endorse or promote
#       products derived from this software without specific prior
#       written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Authors: Stuart Glaser, William Woodall, Robert Haschke
# Maintainer: Morgan Quigley <morgan@osrfoundation.org>

import xml.dom.minidom


def first_child_element(elt):
    c = elt.firstChild
    while c and c.nodeType != xml.dom.Node.ELEMENT_NODE:
        c = c.nextSibling
    return c


def next_sibling_element(node):
    c = node.nextSibling
    while c and c.nodeType != xml.dom.Node.ELEMENT_NODE:
        c = c.nextSibling
    return c


def replace_node(node, by, content_only=False):
    parent = node.parentNode

    if by is not None:
        if not isinstance(by, list):
            by = [by]

        # insert new content before node
        for doc in by:
            if content_only:
                c = doc.firstChild
                while c:
                    n = c.nextSibling
                    parent.insertBefore(c, node)
                    c = n
            else:
                parent.insertBefore(doc, node)

    # remove node
    parent.removeChild(node)


def attribute(tag, a):
    """
    Helper function to fetch a single attribute value from tag
    :param tag (xml.dom.Element): DOM element node
    :param a (str): attribute name
    :return: attribute value if present, otherwise None
    """
    if tag.hasAttribute(a):
        # getAttribute returns empty string for non-existent attributes,
        # which makes it impossible to distinguish with empty values
        return tag.getAttribute(a)
    else:
        return None


def opt_attrs(tag, attrs):
    """
    Helper routine for fetching optional tag attributes
    :param tag (xml.dom.Element): DOM element node
    :param attrs [str]: list of attributes to fetch
    """
    return [attribute(tag, a) for a in attrs]


def reqd_attrs(tag, attrs):
    """
    Helper routine for fetching required tag attributes
    :param tag (xml.dom.Element): DOM element node
    :param attrs [str]: list of attributes to fetch
    :raise RuntimeError: if required attribute is missing
    """
    result = opt_attrs(tag, attrs)
    for (res, name) in zip(result, attrs):
        if res is None:
            raise RuntimeError("%s: missing attribute '%s'" % (tag.nodeName, name))   # pragma: no cover # noqa
    return result


# Better pretty printing of xml
# Taken from
# http://ronrothman.com/public/leftbraned/xml-dom-minidom-toprettyxml-and-silly-whitespace/ # noqa
def fixed_writexml(self, writer, indent="", addindent="", newl=""):   # pragma: no cover # noqa
    # indent = current indentation
    # addindent = indentation to add to higher levels
    # newl = newline string
    writer.write(indent + "<" + self.tagName)

    attrs = self._get_attributes()
    a_names = sorted(attrs.keys())

    for a_name in a_names:
        writer.write(" %s=\"" % a_name)
        xml.dom.minidom._write_data(writer, attrs[a_name].value)
        writer.write("\"")
    if self.childNodes:
        if len(self.childNodes) == 1 \
           and self.childNodes[0].nodeType == xml.dom.minidom.Node.TEXT_NODE:
            writer.write(">")
            self.childNodes[0].writexml(writer, "", "", "")
            writer.write("</%s>%s" % (self.tagName, newl))
            return
        writer.write(">%s" % newl)
        for node in self.childNodes:
            # skip whitespace-only text nodes
            if node.nodeType == xml.dom.minidom.Node.TEXT_NODE and \
                    (not node.data or node.data.isspace()):
                continue
            node.writexml(writer, indent + addindent, addindent, newl)
        writer.write("%s</%s>%s" % (indent, self.tagName, newl))
    else:
        writer.write("/>%s" % newl)


# replace minidom's function with ours
xml.dom.minidom.Element.writexml = fixed_writexml



from roboticstoolbox.tools.null import null
from roboticstoolbox.tools.p_servo import p_servo
from roboticstoolbox.tools.Ticker import Ticker
from roboticstoolbox.tools.urdf import *  # noqa
from roboticstoolbox.tools.trajectory import tpoly, \
    jtraj, ctraj, lspb, t1plot, qplot, mstraj
from roboticstoolbox.tools.numerical import jacobian_numerical, \
    hessian_numerical
from roboticstoolbox.tools.jsingu import jsingu
from roboticstoolbox.tools.data import loaddata, loadmat, path_to_datafile


__all__ = [
    'null',
    'p_servo',
    'Ticker',
    'tpoly',
    'jtraj',
    'ctraj',
    'lspb',
    't1plot',
    'qplot',
    'mstraj',
    'jsingu',
    'jacobian_numerical',
    'hessian_numerical',
    'loaddata',
    'loadmat',
    'path_to_datafile',
]



import copy
from abc import ABC
import numpy as np
from functools import wraps
from spatialmath.base import getvector, isscalar, isvector, ismatrix
from ansitable import ANSITable, Column


def _listen_dyn(func):
    """
    Decorator for property setters

    Use this decorator for any property setter that updates a parameter that
    affects the result of inverse dynamics.  This allows the C version of the
    parameters only having to be updated when they change, rather than on
    every call.  This decorator signals the change by invoking the
    ``dynchanged()`` method of the robot that owns the link.

    Example::

        @m.setter
        @_listen_dyn
        def m(self, m_new):
            self._m = m_new

    """
    @wraps(func)
    def wrapper_listen_dyn(*args):
        if args[0]._robot is not None:
            args[0]._robot.dynchanged()
        return func(*args)
    return wrapper_listen_dyn


class Link(ABC):
    """
    Link superclass

    :param name: name of the link
    :type name: str

    :param qlim: joint variable limits [min, max]
    :type qlim: float ndarray(1,2)
    :param flip: joint moves in opposite direction
    :type flip: bool

    :param m: dynamic - link mass
    :type m: float
    :param r: dynamic - position of COM with respect to link frame
    :type r:  float ndarray(3)
    :param I: dynamic - inertia of link with respect to COM
    :type I: ndarray

    :param Jm: dynamic - motor inertia
    :type Jm: float
    :param B: dynamic - motor viscous friction
    :type B: float, or ndarray(2,)
    :param Tc: dynamic - motor Coulomb friction [Tc⁺, Tc⁻]
    :type Tc: ndarray(2,)
    :param G: dynamic - gear ratio
    :type G: float

    An abstract link superclass for all link types.

    .. inheritance-diagram:: roboticstoolbox.RevoluteDH
        roboticstoolbox.PrismaticDH roboticstoolbox.RevoluteMDH
        roboticstoolbox.PrismaticMDH roboticstoolbox.ELink
        :top-classes: roboticstoolbox.robot.Link
        :parts: 2

    It holds metadata related to:

    - a robot link, such as rigid-body inertial parameters defined in the link
      frame, and link name
    - a robot joint, that connects this link to its parent, such as joint
      limits, direction of motion, motor and transmission parameters.

    .. note::
        - For a more sophisticated actuator model use the ``actuator``
          attribute which is not initialized or used by this Toolbox.
        - There is no ability to name a joint as supported by URDF

    """

    def __init__(
            self,
            name=None,
            qlim=None,
            flip=False,
            m=0.0,
            r=np.zeros((3,)),
            I=np.zeros((3, 3)),  # noqa
            Jm=0.0,
            B=0.0,
            Tc=np.zeros((2,)),
            G=0.0,
            mesh=None,
            geometry=[],
            collision=[],
            **kwargs):

        self._robot = None  # reference to owning robot

        self._name = name

        # Joint angle parameters
        self.flip = flip
        self.qlim = qlim

        # Link geometry
        self.geometry = geometry
        self.collision = collision

        # TODO this is a leftover from VPython development, should be
        # using geometry instead
        self.mesh = mesh

        # Link dynamic Parameters
        self.m = m
        self.r = r
        self.I = I  # noqa

        # Motor dynamic parameters
        self.Jm = Jm
        self.B = B
        self.Tc = Tc
        self.G = G
        self.actuator = None  # reference to more advanced actuator model

    def copy(self):
        """
        Copy of link object

        :return: Shallow copy of link object
        :rtype: Link

        ``link.copy()`` is a new Link subclass instance with a copy of all
        the parameters.
        """
        new = copy.copy(self)
        for k, v in self.__dict__.items():
            if k.startswith('_') and isinstance(v, np.ndarray):
                setattr(new, k, np.copy(v))
        return new

    def _copy(self):
        raise DeprecationWarning('Use copy method of Link class')

    def dyn(self, indent=0):
        """
        Inertial properties of link as a string

        :param indent: indent each line by this many spaces
        :type indent: int
        :return: The string representation of the link dynamics
        :rtype: string

        ``link.dyn()`` is a string representation the inertial properties of
        the link object in a multi-line format. The properties shown are mass,
        centre of mass, inertia, friction, gear ratio and motor properties.

        Example:

        .. runblock:: pycon

            >>> import roboticstoolbox as rtb
            >>> robot = rtb.models.DH.Puma560()
            >>> print(robot.links[2])        # kinematic parameters
            >>> print(robot.links[2].dyn())  # dynamic parameters

        :seealso: :func:`~dyntable`
        """

        s = "m     =  {:8.2g} \n" \
            "r     =  {:8.2g} {:8.2g} {:8.2g} \n" \
            "        | {:8.2g} {:8.2g} {:8.2g} | \n" \
            "I     = | {:8.2g} {:8.2g} {:8.2g} | \n" \
            "        | {:8.2g} {:8.2g} {:8.2g} | \n" \
            "Jm    =  {:8.2g} \n" \
            "B     =  {:8.2g} \n" \
            "Tc    =  {:8.2g}(+) {:8.2g}(-) \n" \
            "G     =  {:8.2g} \n" \
            "qlim  =  {:8.2g} to {:8.2g}".format(
                self.m,
                self.r[0], self.r[1], self.r[2],
                self.I[0, 0], self.I[0, 1], self.I[0, 2],
                self.I[1, 0], self.I[1, 1], self.I[1, 2],
                self.I[2, 0], self.I[2, 1], self.I[2, 2],
                self.Jm,
                self.B,
                self.Tc[0], self.Tc[1],
                self.G,
                self.qlim[0], self.qlim[1]
            )

        if indent > 0:
            # insert indentations into the string
            # TODO there is probably a tidier way to integrate this step with
            # above
            sp = ' ' * indent
            s = sp + s.replace('\n', '\n' + sp)

        return s

    def _dyn2list(self, fmt="{: .3g}"):
        """
        Inertial properties of link as a string

        :param fmt: conversion format for each number
        :type fmt: str
        :return: The string representation of the link dynamics
        :rtype: string

        ``link.)_dyn2list()`` returns a list of pretty-printed inertial
        properties of the link The properties included are mass, centre of
        mass, inertia, friction, gear ratio and motor properties.

        :seealso: :func:`~dyn`
        """
        ANSITable(
            Column("Parameter", headalign="^"),
            Column("Value", headalign="^", colalign="<"),
            border="thin")

        def format(l, fmt, val):  # noqa
            if isinstance(val, np.ndarray):
                s = ', '.join([fmt.format(v) for v in val])
            else:
                s = fmt.format(val)
            l.append(s)

        dyn = []
        format(dyn, fmt, self.m)
        format(dyn, fmt, self.r)
        I = self.I.flatten()  # noqa
        format(dyn, fmt, np.r_[[I[k] for k in [0, 4, 8, 1, 2, 5]]])
        format(dyn, fmt, self.Jm)
        format(dyn, fmt, self.B)
        format(dyn, fmt, self.Tc)
        format(dyn, fmt, self.G)

        return dyn

    def _format(self, l, name, ignorevalue=0, indices=None):  # noqa  # pragma nocover
        v = getattr(self, name)
        s = None
        if v is None:
            return
        if isscalar(v) and v != ignorevalue:
            s = f"{name}={v}"
        elif isinstance(v, np.ndarray):
            if np.linalg.norm(v, ord=np.inf) > 0:
                if indices is not None:
                    flat = v.flatten()
                    v = np.r_[[flat[k] for k in indices]]
                s = f"{name}=[" + ", ".join([str(x) for x in v]) + "]"
        if s is not None:
            l.append(s)

    def _params(self):  # pragma nocover
        l = []  # noqa
        self._format(l, "name")
        self._format(l, "flip", False)
        self._format(l, "qlim")
        self._format(l, "m")
        self._format(l, "r")
        self._format(l, "I", indices=[0, 4, 8, 1, 2, 5])
        self._format(l, "Jm")
        self._format(l, "B")
        self._format(l, "Tc")
        self._format(l, "G")

        return l

    def islimit(self, q):
        """
        Checks if joint exceeds limit

        :param q: joint coordinate
        :type q: float
        :return: True if joint is exceeded
        :rtype: bool

        ``link.islimit(q)`` is True if ``q`` exceeds the joint limits defined
        by ``link``.

        .. note:: If no limits are set always return False.

        :seealso: :func:`qlim`
        """
        if self.qlim is None:
            return False
        else:
            return q < self.qlim[0] or q > self.qlim[1]

    def nofriction(self, coulomb=True, viscous=False):
        """
        Clone link without friction

        :param coulomb: if True, will set the Coulomb friction to 0
        :type coulomb: bool
        :param viscous: if True, will set the viscous friction to 0
        :type viscous: bool

        ``link.nofriction()`` is a copy of the link instance with the same
        parameters except, the Coulomb and/or viscous friction parameters are
        set to zero.

        .. note:: For simulation it can be useful to remove Couloumb friction
            which can cause problems for numerical integration.

        """

        # Copy the Link
        link = self.copy()

        if viscous:
            link.B = 0.0

        if coulomb:
            link.Tc = [0.0, 0.0]

        return link

    def friction(self, qd, coulomb=True):
        r"""
        Compute joint friction

        :param qd: The joint velocity
        :type qd: float
        :param coulomb: include Coulomb friction
        :type coloumb: bool, default True
        :return tau: the friction force/torque
        :rtype tau: float

        ``friction(qd)`` is the joint friction force/torque
        for joint velocity ``qd``. The friction model includes:

        - Viscous friction which is a linear function of velocity.
        - Coulomb friction which is proportional to sign(qd).

        .. math::

            \tau = G^2 B \dot{q} + |G| \left\{ \begin{array}{ll}
                \tau_C^+ & \mbox{if $\dot{q} > 0$} \\
                \tau_C^- & \mbox{if $\dot{q} < 0$} \end{array} \right.

        .. note::

            - The friction value should be added to the motor output torque to
              determine the nett torque. It has a negative value when qd > 0.
            - The returned friction value is referred to the output of the
              gearbox.
            - The friction parameters in the Link object are referred to the
              motor.
            - Motor viscous friction is scaled up by :math:`G^2`.
            - Motor Coulomb friction is scaled up by math:`G`.
            - The appropriate Coulomb friction value to use in the
              non-symmetric case depends on the sign of the joint velocity,
              not the motor velocity.
            - Coulomb friction is zero for zero joint velocity, stiction is
              not modeled.
            - The absolute value of the gear ratio is used.  Negative gear
              ratios are tricky: the Puma560 robot has negative gear ratio for
              joints 1 and 3.

        """

        tau = self.B * np.abs(self.G) * qd

        if coulomb:
            if qd > 0:
                tau += self.Tc[0]
            elif qd < 0:
                tau += self.Tc[1]

        # Scale up by gear ratio
        tau = -np.abs(self.G) * tau

        return tau

# -------------------------------------------------------------------------- #

    @property
    def name(self):
        """
        Get/set link name

        - ``link.name`` is the link name

        :return: link name
        :rtype: str

        - ``link.name = ...`` checks and sets the link name
        """
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

# -------------------------------------------------------------------------- #

    @property
    def qlim(self):
        """
        Get/set joint limits

        - ``link.qlim`` is the joint limits

        :return: joint limits
        :rtype: ndarray(2,) or Nine

        - ``link.qlim = ...`` checks and sets the joint limits

        .. note::
            - The limits are not widely enforced within the toolbox.
            - If no joint limits are specified the value is ``None``

        :seealso: :func:`~islimit`
        """
        return self._qlim

    @qlim.setter
    def qlim(self, qlim_new):
        if qlim_new is None:
            self._qlim = None
        else:
            self._qlim = getvector(qlim_new, 2)

# -------------------------------------------------------------------------- #

    @property
    def flip(self):
        """
        Get/set joint flip

        - ``link.flip`` is the joint flip status

        :return: joint flip
        :rtype: bool

        - ``link.flip = ...`` checks and sets the joint flip status

        Joint flip defines the direction of motion of the joint.

        ``flip = False`` is conventional motion direction:

            - revolute motion is a positive rotation about the z-axis
            - prismatic motion is a positive translation along the z-axis

        ``flip = True`` is the opposite motion direction:

            - revolute motion is a negative rotation about the z-axis
            - prismatic motion is a negative translation along the z-axis

        """
        return self._flip

    @flip.setter
    def flip(self, flip_new):
        self._flip = flip_new

# -------------------------------------------------------------------------- #

    @property
    def m(self):
        """
        Get/set link mass

        - ``link.m`` is the link mass

        :return: link mass
        :rtype: float

        - ``link.m = ...`` checks and sets the link mass

        """
        return self._m

    @m.setter
    @_listen_dyn
    def m(self, m_new):
        self._m = m_new

# -------------------------------------------------------------------------- #

    @property
    def r(self):
        """
        Get/set link centre of mass

        - ``link.r`` is the link centre of mass

        :return: link centre of mass
        :rtype: ndarray(3,)

        - ``link.r = ...`` checks and sets the link centre of mass

        The link centre of mass is a 3-vector defined with respect to the link
        frame.
        """
        return self._r

    @r.setter
    @_listen_dyn
    def r(self, r_new):
        self._r = getvector(r_new, 3)

# -------------------------------------------------------------------------- #

    @property
    def I(self):    # noqa
        r"""
        Get/set link inertia

        - ``link.I`` is the link inertia

        :return: link inertia
        :rtype: ndarray(3,3)

        - ``link.I = ...`` checks and sets the link inertia

        Link inertia is a symmetric 3x3 matrix describing the inertia with
        respect to a frame with its origin at the centre of mass, and with
        axes parallel to those of the link frame.

        The inertia matrix is

        :math:`\begin{bmatrix} I_{xx} & I_{xy} & I_{xz} \\ I_{xy} & I_{yy} & I_{yz} \\I_{xz} & I_{yz} & I_{zz} \end{bmatrix}`


        and can be specified as either:

        - a 3 ⨉ 3 symmetric matrix
        - a 3-vector :math:`(I_{xx}, I_{yy}, I_{zz})`
        - a 6-vector :math:`(I_{xx}, I_{yy}, I_{zz}, I_{xy}, I_{yz}, I_{xz})`

        .. note:: Referred to the link side of the gearbox.
        """  # noqa
        return self._I

    @I.setter
    @_listen_dyn
    def I(self, I_new):  # noqa

        if ismatrix(I_new, (3, 3)):
            # 3x3 matrix passed
            if np.any(np.abs(I_new - I_new.T) > 1e-8):
                raise ValueError('3x3 matrix is not symmetric')

        elif isvector(I_new, 9):
            # 3x3 matrix passed as a 1d vector
            I_new = I_new.reshape(3, 3)
            if np.any(np.abs(I_new - I_new.T) > 1e-8):
                raise ValueError('3x3 matrix is not symmetric')

        elif isvector(I_new, 6):
            # 6-vector passed, moments and products of inertia,
            # [Ixx Iyy Izz Ixy Iyz Ixz]
            I_new = np.array([
                [I_new[0], I_new[3], I_new[5]],
                [I_new[3], I_new[1], I_new[4]],
                [I_new[5], I_new[4], I_new[2]]
            ])

        elif isvector(I_new, 3):
            # 3-vector passed, moments of inertia [Ixx Iyy Izz]
            I_new = np.diag(I_new)

        else:
            raise ValueError('invalid shape passed: must be (3,3), (6,), (3,)')

        self._I = I_new

# -------------------------------------------------------------------------- #

    @property
    def Jm(self):
        """
        Get/set motor inertia

        - ``link.Jm`` is the motor inertia

        :return: motor inertia
        :rtype: float

        - ``link.Jm = ...`` checks and sets the motor inertia

        .. note:: Referred to the motor side of the gearbox.
        """
        return self._Jm

    @Jm.setter
    @_listen_dyn
    def Jm(self, Jm_new):
        self._Jm = Jm_new

# -------------------------------------------------------------------------- #

    @property
    def B(self):
        """
        Get/set motor viscous friction

        - ``link.B`` is the motor viscous friction

        :return: motor viscous friction
        :rtype: float

        - ``link.B = ...`` checks and sets the motor viscous friction

        .. note::
            - Referred to the motor side of the gearbox.
            - Viscous friction is the same for positive and negative motion.
        """
        return self._B

    @B.setter
    @_listen_dyn
    def B(self, B_new):
        if isscalar(B_new):
            self._B = B_new
        else:
            raise TypeError("B must be a scalar")

# -------------------------------------------------------------------------- #

    @property
    def Tc(self):
        r"""
        Get/set motor Coulomb friction

        - ``link.Tc`` is the motor Coulomb friction

        :return: motor Coulomb friction
        :rtype: ndarray(2)

        - ``link.Tc = ...`` checks and sets the motor Coulomb friction. If a
          scalar is given the value is set to [T, -T], if a 2-vector it is
          assumed to be in the order [Tc⁺, Tc⁻]

        Coulomb friction is a non-linear friction effect defined by two
        parameters such that

        .. math::

            \tau = \left\{ \begin{array}{ll}
                \tau_C^+ & \mbox{if $\dot{q} > 0$} \\
                \tau_C^- & \mbox{if $\dot{q} < 0$} \end{array} \right.

        .. note::
            -  Referred to the motor side of the gearbox.
            - :math:`\tau_C^+` must be :math:`> 0`, and :math:`\tau_C^-` must
              be :math:`< 0`.
        """
        return self._Tc

    @Tc.setter
    @_listen_dyn
    def Tc(self, Tc_new):

        try:
            # sets Coulomb friction parameters to [F -F], for a symmetric
            # Coulomb friction model.
            Tc = getvector(Tc_new, 1)
            Tc_new = np.array([Tc[0], -Tc[0]])
        except ValueError:
            # [FP FM] sets Coulomb friction to [FP FM], for an asymmetric
            # Coulomb friction model. FP>0 and FM<0.  FP is applied for a
            # positive joint velocity and FM for a negative joint
            # velocity.
            Tc_new = getvector(Tc_new, 2)

        self._Tc = Tc_new

# -------------------------------------------------------------------------- #

    @property
    def G(self):
        """
        Get/set gear ratio

        - ``link.G`` is the transmission gear ratio

        :return: gear ratio
        :rtype: float

        - ``link.G = ...`` checks and sets the gear ratio

        .. note::
            - The ratio of motor motion : link motion
            - The gear ratio can be negative, see also the ``flip`` attribute.

        :seealso: :func:`flip`
        """
        return self._G

    @G.setter
    @_listen_dyn
    def G(self, G_new):
        self._G = G_new

# -------------------------------------------------------------------------- #

    def closest_point(self, shape, inf_dist=1.0):
        '''
        closest_point(shape, inf_dist) returns the minimum euclidean
        distance between this link and shape, provided it is less than
        inf_dist. It will also return the points on self and shape in the
        world frame which connect the line of length distance between the
        shapes. If the distance is negative then the shapes are collided.

        :param shape: The shape to compare distance to
        :type shape: Shape
        :param inf_dist: The minimum distance within which to consider
            the shape
        :type inf_dist: float

        :returns: d, p1, p2 where d is the distance between the shapes,
            p1 and p2 are the points in the world frame on the respective
            shapes
        :rtype: float, SE3, SE3
        '''

        d = 10000
        p1 = None,
        p2 = None

        for col in self.collision:
            td, tp1, tp2 = col.closest_point(shape, inf_dist)

            if td is not None and td < d:
                d = td
                p1 = tp1
                p2 = tp2

        if d == 10000:
            d = None

        return d, p1, p2

    def collided(self, shape):
        '''
        collided(shape) checks if this link and shape have collided

        :param shape: The shape to compare distance to
        :type shape: Shape

        :returns: True if shapes have collided
        :rtype: bool
        '''

        for col in self.collision:
            if col.collided(shape):
                return True

        return False


if __name__ == "__main__":  # pragma nocover

    import roboticstoolbox as rtb
    from spatialmath.base import sym

    ET = rtb.ETS

    d, a, theta = sym.symbol('d, a, theta')

    e = ET.tx(d) * ET.ty(a) * ET.rz(theta)
    print(e)
    link = rtb.ELink()



#!/usr/bin/env python

import numpy as np
from spatialmath import SE3


def p_servo(wTe, wTep, gain=2, threshold=0.1):
    '''
    Position-based servoing.

    Returns the end-effector velocity which will cause the robot to approach
    the desired pose.

    :param wTe: The current pose of the end-effecor in the base frame.
    :type wTe: SE3
    :param wTep: The desired pose of the end-effecor in the base frame.
    :type wTep: SE3
    :param gain: The gain for the controller
    :type gain: float
    :param threshold: The threshold or tolerance of the final error between
        the robot's pose and desired pose
    :type threshold: float

    :returns v: The velocity of the end-effecotr which will casue the robot
        to approach wTep
    :rtype v: ndarray(6)
    :returns arrived: True if the robot is within the threshold of the final
        pose
    :rtype arrived: bool

    '''

    if not isinstance(wTe, SE3):
        wTe = SE3(wTe)

    if not isinstance(wTep, SE3):
        wTep = SE3(wTep)

    # Pose difference
    eTep = wTe.inv() * wTep

    # Translational velocity error
    ev = eTep.t

    # Angular velocity error
    ew = eTep.rpy('rad')

    # Form error vector
    e = np.r_[ev, ew]

    # Desired end-effector velocity
    v = gain * e

    if np.sum(np.abs(e)) < threshold:
        arrived = True
    else:
        arrived = False

    return v, arrived



#!/usr/bin/env python
"""
@author: Jesse Haviland
"""

import numpy as np
from spatialmath import SE3
import roboticstoolbox as rp
from roboticstoolbox.robot.Link import Link, _listen_dyn
from roboticstoolbox.robot.ETS import ETS

_eps = np.finfo(np.float64).eps

from functools import wraps
def _check_rne(func):
    @wraps(func)
    def wrapper_check_rne(*args, **kwargs):
        if args[0]._rne_ob is None or args[0]._dynchanged:
            args[0].delete_rne()
            args[0]._init_rne()
        args[0]._rne_changed = False
        return func(*args, **kwargs)
    return wrapper_check_rne

# --------------------------------------------------------------#

try:  # pragma: no cover
    # print('Using SymPy')
    import sympy as sym

    def _issymbol(x):
        return isinstance(x, sym.Expr)
except ImportError:
    def _issymbol(x):  # pylint: disable=unused-argument
        return False


def _cos(theta):
    if _issymbol(theta):
        return sym.cos(theta)
    else:
        return np.cos(theta)


def _sin(theta):
    if _issymbol(theta):
        return sym.sin(theta)
    else:
        return np.sin(theta)

# --------------------------------------------------------------#


class DHLink(Link):
    """
    A link superclass for all robots defined using Denavit-Hartenberg notation.
    A Link object holds all information related to a robot joint and link such
    as kinematics parameters, rigid-body inertial parameters, motor and
    transmission parameters.

    :param theta: kinematic: joint angle
    :type theta: float
    :param d: kinematic - link offset
    :type d: float
    :param alpha: kinematic - link twist
    :type alpha: float
    :param a: kinematic - link length
    :type a: float
    :param sigma: kinematic - 0 if revolute, 1 if prismatic
    :type sigma: int
    :param mdh: kinematic - 0 if standard D&H, else 1
    :type mdh: int
    :param offset: kinematic - joint variable offset
    :type offset: float

    :param qlim: joint variable limits [min, max]
    :type qlim: ndarray(2,)
    :param flip: joint moves in opposite direction
    :type flip: bool

    :param m: dynamic - link mass
    :type m: float
    :param r: dynamic - position of COM with respect to link frame
    :type r:  ndarray(3,)
    :param I: dynamic - inertia of link with respect to COM
    :type I: ndarray
    :param Jm: dynamic - motor inertia
    :type Jm: float
    :param B: dynamic - motor viscous friction: B=B⁺=B⁻, [B⁺, B⁻]
    :type B: float, or ndarray(2,)
    :param Tc: dynamic - motor Coulomb friction [Tc⁺, Tc⁻]
    :type Tc: ndarray(2,)
    :param G: dynamic - gear ratio
    :type G: float

    :references:
        - Robotics, Vision & Control, P. Corke, Springer 2011, Chap 7.

    """

    def __init__(
            self,
            d=0.0,
            alpha=0.0,
            theta=0.0,
            a=0.0,
            sigma=0,
            mdh=False,
            offset=0,
            **kwargs):

        # TODO
        #  probably should make DHLink(link) return a copy
        #  probably should enforce keyword arguments, easy to make an
        #    error with positional args
        super().__init__(**kwargs)

        # Kinematic parameters
        self.sigma = sigma
        self.theta = theta
        self.d = d
        self.alpha = alpha
        self.a = a
        self.mdh = mdh
        self.offset = offset
        self.id = None

    def __add__(self, L):
        if isinstance(L, DHLink):
            return rp.DHRobot([self, L])

        elif isinstance(L, rp.DHRobot):
            nlinks = [self]

            # TODO - Should I do a deep copy here a physically copy the Links
            # and not just the references?
            # Copy Link references to new list
            for i in range(L.n):
                nlinks.append(L.links[i])

            return rp.DHRobot(
                nlinks,
                name=L.name,
                manufacturer=L.manufacturer,
                base=L.base,
                tool=L.tool,
                gravity=L.gravity)

        else:
            raise TypeError("Cannot add a Link with a non Link object")

    def __str__(self):

        if self.offset == 0:
            offset = ""
        else:
            offset = f" + {self.offset}"
        if self.id is None:
            qvar = "q"
        else:
            qvar = f"q{self.id}"
        cls = self.__class__.__name__
        if self.isrevolute:
            s = f"{cls}:   theta={qvar}{offset},  d={self.d}, " \
                f"a={self.a}, alpha={self.alpha}"
        elif self.isprismatic:
            s = f"{cls}:  theta={self.theta},  d={qvar}{offset}, " \
                f" a={self.a},  " \
                f"alpha={self.alpha}"
        return s

    def __repr__(self):
        name = self.__class__.__name__
        args = []
        if self.isrevolute:
            self._format(args, "d")
        else:
            self._format(args, "theta")
        self._format(args, "a")
        self._format(args, "alpha")
        args.extend(super()._params())
        return name + "(" + ", ".join(args) + ")"

# -------------------------------------------------------------------------- #

    @property
    def theta(self):
        """
        Get/set joint angle

        - ``link.theta`` is the joint angle
            :return: joint angle
            :rtype: float
        - ``link.theta = ...`` checks and sets the joint angle

        """
        return self._theta

    @theta.setter
    @_listen_dyn
    def theta(self, theta_new):
        if not self.sigma and theta_new != 0.0:
            raise ValueError("theta is not valid for revolute joints")
        else:
            self._theta = theta_new

# -------------------------------------------------------------------------- #

    @property
    def d(self):
        """
        Get/set link offset

        - ``link.d`` is the link offset
            :return: link offset
            :rtype: float
        - ``link.d = ...`` checks and sets the link offset

        """
        return self._d

    @d.setter
    @_listen_dyn
    def d(self, d_new):
        if self.sigma and d_new != 0.0:
            raise ValueError("f is not valid for prismatic joints")
        else:
            self._d = d_new

# -------------------------------------------------------------------------- #

    @property
    def a(self):
        """
        Get/set link length

        - ``link.a`` is the link length
            :return: link length
            :rtype: float
        - ``link.a = ...`` checks and sets the link length

        """
        return self._a

    @a.setter
    @_listen_dyn
    def a(self, a_new):
        self._a = a_new
# -------------------------------------------------------------------------- #

    @property
    def alpha(self):
        """
        Get/set link twist

        - ``link.d`` is the link twist
            :return: link twist
            :rtype: float
        - ``link.d = ...`` checks and sets the link twist

        """
        return self._alpha

    @alpha.setter
    @_listen_dyn
    def alpha(self, alpha_new):
        self._alpha = alpha_new

# -------------------------------------------------------------------------- #

    @property
    def sigma(self):
        """
        Get/set joint type

        - ``link.sigma`` is the joint type
            :return: joint type
            :rtype: int
        - ``link.sigma = ...`` checks and sets the joint type

        The joint type is 0 for a revolute joint, and 1 for a prismatic joint.

        :seealso: :func:`isrevolute`, :func:`isprismatic`
        """
        return self._sigma

    @sigma.setter
    @_listen_dyn
    def sigma(self, sigma_new):
        self._sigma = sigma_new
# -------------------------------------------------------------------------- #

    @property
    def mdh(self):
        """
        Get/set kinematic convention

        - ``link.mdh`` is the kinematic convention
            :return: kinematic convention
            :rtype: bool
        - ``link.mdh = ...`` checks and sets the kinematic convention

        The kinematic convention is True for modified Denavit-Hartenberg
        notation (eg. Craig's textbook) and False for Denavit-Hartenberg
        notation (eg. Siciliano, Spong, Paul textbooks).
        """
        return self._mdh

    @mdh.setter
    @_listen_dyn
    def mdh(self, mdh_new):
        self._mdh = int(mdh_new)

# -------------------------------------------------------------------------- #

    @property
    def offset(self):
        """
        Get/set joint variable offset

        - ``link.offset`` is the joint variable offset

        :return: joint variable offset
        :rtype: float

        - ``link.offset = ...`` checks and sets the joint variable offset

        The offset is added to the joint angle before forward kinematics, and
        subtracted after inverse kinematics.  It is used to define the joint
        configuration for zero joint coordinates.

        """
        return self._offset

    @offset.setter
    def offset(self, offset_new):
        self._offset = offset_new

# -------------------------------------------------------------------------- #

    def A(self, q):
        r"""
        Link transform matrix

        :param q: Joint coordinate
        :type q: float
        :return T: SE(3) link homogeneous transformation
        :rtype T: SE3 instance

        ``A(q)`` is an ``SE3`` instance representing the SE(3) homogeneous
        transformation matrix corresponding to the link's joint variable ``q``
        which is either the Denavit-Hartenberg parameter :math:`\theta_j`
        (revolute) or :math:`d_j` (prismatic).

        This is the relative pose of the current link frame with respect to the
        previous link frame.

        For details of the computation see the documentation for the
        subclasses, click on the right side of the class boxes below.

        .. inheritance-diagram:: roboticstoolbox.RevoluteDH
            roboticstoolbox.PrismaticDH roboticstoolbox.RevoluteMDH
            roboticstoolbox.PrismaticMDH
            :top-classes: roboticstoolbox.robot.DHLink.DHLink
            :parts: 2

        .. note::

            - For a revolute joint the ``theta`` parameter of the link is
              ignored, and ``q`` used instead.
            - For a prismatic joint the ``d`` parameter of the link is ignored,
              and ``q`` used instead.
            - The joint ``offset`` parameter is added to ``q`` before
              computation of the transformation matrix.
            - The computation is different for standard and modified
              Denavit-Hartenberg parameters.

        :seealso: :class:`RevoluteDH`, :class:`PrismaticDH`,
            :class:`RevoluteMDH`, :class:`PrismaticMDH`
        """

        sa = _sin(self.alpha)
        ca = _cos(self.alpha)

        if self.flip:
            q = -q + self.offset
        else:
            q = q + self.offset

        if self.sigma == 0:
            # revolute
            st = _sin(q)
            ct = _cos(q)
            d = self.d
        else:
            # prismatic
            st = _sin(self.theta)
            ct = _cos(self.theta)
            d = q

        if self.mdh == 0:
            # standard DH
            T = np.array([
                [ct, -st * ca, st * sa, self.a * ct],
                [st, ct * ca, -ct * sa, self.a * st],
                [0, sa, ca, d],
                [0, 0, 0, 1]
            ])
        else:
            # modified DH
            T = np.array([
                [ct, -st, 0, self.a],
                [st * ca, ct * ca, -sa, -sa * d],
                [st * sa, ct * sa, ca, ca * d],
                [0, 0, 0, 1]
            ])

        return SE3(T, check=False)

    @property
    def isrevolute(self):
        """
        Checks if the joint is of revolute type

        :return: Ture if is revolute
        :rtype: bool

        :seealso: :func:`sigma`
        """

        if not self.sigma:
            return True
        else:
            return False

    @property
    def isprismatic(self):
        """
        Checks if the joint is of prismatic type

        :return: Ture if is prismatic
        :rtype: bool

        :seealso: :func:`sigma`
        """

        if self.sigma:
            return True
        else:
            return False

    def ets(self):
        ets = ETS()

        if self.mdh:
            # MDH format: a alpha theta d
            if self.a != 0:
                ets *= ETS.tx(self.a)
            if self.alpha != 0:
                ets *= ETS.rx(self.alpha)

            if self.isrevolute:
                if self.offset != 0:
                    ets *= ETS.rz(self.offset)
                ets *= ETS.rz(flip=self.flip)  # joint

                if self.d != 0:
                    ets *= ETS.tz(self.d)
            elif self.isprismatic:
                if self.theta != 0:
                    ets *= ETS.rz(self.theta)

                if self.offset != 0:
                    ets *= ETS.tz(self.offset)
                ets *= ETS.tz(flip=self.flip)  # joint

        else:
            # DH format: theta d a alpha

            if self.isrevolute:
                ets *= ETS.rz(flip=self.flip)
                if self.offset != 0:
                    ets *= ETS.rz(self.offset)

                if self.d != 0:
                    ets *= ETS.tz(self.d)
            elif self.isprismatic:
                if self.theta != 0:
                    ets *= ETS.rz(self.theta)

                if self.offset != 0:
                    ets *= ETS.tz(self.offset)
                ets *= ETS.tz(flip=self.flip)

            if self.a != 0:
                ets *= ETS.tx(self.a)
            if self.alpha != 0:
                ets *= ETS.rx(self.alpha)
        return ets


# -------------------------------------------------------------------------- #

class RevoluteDH(DHLink):
    r"""
    Class for revolute links using standard DH convention

    :param d: kinematic - link offset
    :type d: float
    :param alpha: kinematic - link twist
    :type alpha: float
    :param a: kinematic - link length
    :type a: float
    :param offset: kinematic - joint variable offset
    :type offset: float

    :param qlim: joint variable limits [min, max]
    :type qlim: float ndarray(1,2)
    :param flip: joint moves in opposite direction
    :type flip: bool

    :param m: dynamic - link mass
    :type m: float
    :param r: dynamic - position of COM with respect to link frame
    :type r:  float ndarray(3)
    :param I: dynamic - inertia of link with respect to COM
    :type I: ndarray
    :param Jm: dynamic - motor inertia
    :type Jm: float
    :param B: dynamic - motor viscous friction: B=B⁺=B⁻, [B⁺, B⁻]
    :type B: float, or ndarray(2,)
    :param Tc: dynamic - motor Coulomb friction [Tc⁺, Tc⁻]
    :type Tc: ndarray(2,)
    :param G: dynamic - gear ratio
    :type G: float

    A subclass of the :class:`DHLink` class for a revolute joint that holds all
    information related to a robot link such as kinematics parameters,
    rigid-body inertial parameters, motor and transmission parameters.

    The link transform is

    :math:`\underbrace{\mathbf{T}_{rz}(q_i)}_{\mbox{variable}} \cdot \mathbf{T}_{tz}(d_i) \cdot \mathbf{T}_{tx}(a_i) \cdot \mathbf{T}_{rx}(\alpha_i)`

    where :math:`q_i` is the joint variable.

    :references:
        - Robotics, Vision & Control, P. Corke, Springer 2011, Chap 7.

    :seealso: :func:`PrismaticDH`, :func:`DHLink`, :func:`RevoluteMDH`
    """  # noqa

    def __init__(
            self,
            d=0.0,
            a=0.0,
            alpha=0.0,
            offset=0.0,
            qlim=None,
            flip=False,
            **kwargs
            ):

        theta = 0.0
        sigma = 0
        mdh = False

        super().__init__(
            d=d, alpha=alpha, theta=theta, a=a, sigma=sigma, mdh=mdh,
            offset=offset, qlim=qlim, flip=flip, **kwargs)


class PrismaticDH(DHLink):
    r"""
    Class for prismatic link using standard DH convention

    :param theta: kinematic: joint angle
    :type theta: float
    :param d: kinematic - link offset
    :type d: float
    :param alpha: kinematic - link twist
    :type alpha: float
    :param a: kinematic - link length
    :type a: float
    :param offset: kinematic - joint variable offset
    :type offset: float

    :param qlim: joint variable limits [min, max]
    :type qlim: float ndarray(1,2)
    :param flip: joint moves in opposite direction
    :type flip: bool

    :param m: dynamic - link mass
    :type m: float
    :param r: dynamic - position of COM with respect to link frame
    :type r:  float ndarray(3)
    :param I: dynamic - inertia of link with respect to COM
    :type I: ndarray
    :param Jm: dynamic - motor inertia
    :type Jm: float
    :param B: dynamic - motor viscous friction: B=B⁺=B⁻, [B⁺, B⁻]
    :type B: float, or ndarray(2,)
    :param Tc: dynamic - motor Coulomb friction [Tc⁺, Tc⁻]
    :type Tc: ndarray(2,)
    :param G: dynamic - gear ratio
    :type G: float

    A subclass of the DHLink class for a prismatic joint that holds all
    information related to a robot link such as kinematics parameters,
    rigid-body inertial parameters, motor and transmission parameters.

    The link transform is

    :math:`\mathbf{T}_{rz}(\theta_i) \cdot \underbrace{\mathbf{T}_{tz}(q_i)}_{\mbox{variable}} \cdot \mathbf{T}_{tx}(a_i) \cdot \mathbf{T}_{rx}(\alpha_i)`

    where :math:`q_i` is the joint variable.

    :references:
        - Robotics, Vision & Control, P. Corke, Springer 2011, Chap 7.

    :seealso: :func:`RevoluteDH`, :func:`DHLink`, :func:`PrismaticMDH`
    """  # noqa

    def __init__(
            self,
            theta=0.0,
            a=0.0,
            alpha=0.0,
            offset=0.0,
            qlim=None,
            flip=False,
            **kwargs
            ):

        d = 0.0
        sigma = 1
        mdh = False

        super().__init__(
            theta=theta, d=d, a=a, alpha=alpha, sigma=sigma, mdh=mdh,
            offset=offset, qlim=qlim, flip=flip, **kwargs)


class RevoluteMDH(DHLink):
    r"""
    Class for revolute links using modified DH convention

    :param d: kinematic - link offset
    :type d: float
    :param alpha: kinematic - link twist
    :type alpha: float
    :param a: kinematic - link length
    :type a: float
    :param offset: kinematic - joint variable offset
    :type offset: float

    :param qlim: joint variable limits [min, max]
    :type qlim: float ndarray(1,2)
    :param flip: joint moves in opposite direction
    :type flip: bool

    :param m: dynamic - link mass
    :type m: float
    :param r: dynamic - position of COM with respect to link frame
    :type r:  float ndarray(3)
    :param I: dynamic - inertia of link with respect to COM
    :type I: ndarray
    :param Jm: dynamic - motor inertia
    :type Jm: float
    :param B: dynamic - motor viscous friction: B=B⁺=B⁻, [B⁺, B⁻]
    :type B: float, or ndarray(2,)
    :param Tc: dynamic - motor Coulomb friction [Tc⁺, Tc⁻]
    :type Tc: ndarray(2,)
    :param G: dynamic - gear ratio
    :type G: float

    A subclass of the DHLink class for a revolute joint that holds all
    information related to a robot link such as kinematics parameters,
    rigid-body inertial parameters, motor and transmission parameters.

    The link transform is

    :math:`\mathbf{T}_{tx}(a_{i-1}) \cdot \mathbf{T}_{rx}(\alpha_{i-1}) \cdot \underbrace{\mathbf{T}_{rz}(q_i)}_{\mbox{variable}} \cdot \mathbf{T}_{tz}(d_i)`

    where :math:`q_i` is the joint variable.

    :references:
        - Robotics, Vision & Control, P. Corke, Springer 2011, Chap 7.

    :seealso: :func:`PrismaticMDH`, :func:`DHLink`, :func:`RevoluteDH`
    """  # noqa

    def __init__(
            self,
            d=0.0,
            a=0.0,
            alpha=0.0,
            offset=0.0,
            qlim=None,
            flip=False,
            **kwargs
            ):

        theta = 0.0
        sigma = 0
        mdh = True

        super().__init__(
            d=d, alpha=alpha, theta=theta, a=a, sigma=sigma, mdh=mdh,
            offset=offset, qlim=qlim, flip=flip, **kwargs)


class PrismaticMDH(DHLink):
    r"""
    Class for prismatic link using modified DH convention

    :param theta: kinematic: joint angle
    :type theta: float
    :param d: kinematic - link offset
    :type d: float
    :param alpha: kinematic - link twist
    :type alpha: float
    :param a: kinematic - link length
    :type a: float
    :param offset: kinematic - joint variable offset
    :type offset: float

    :param qlim: joint variable limits [min, max]
    :type qlim: float ndarray(1,2)
    :param flip: joint moves in opposite direction
    :type flip: bool

    :param m: dynamic - link mass
    :type m: float
    :param r: dynamic - position of COM with respect to link frame
    :type r:  float ndarray(3)
    :param I: dynamic - inertia of link with respect to COM
    :type I: ndarray
    :param Jm: dynamic - motor inertia
    :type Jm: float
    :param B: dynamic - motor viscous friction: B=B⁺=B⁻, [B⁺, B⁻]
    :type B: float, or ndarray(2,)
    :param Tc: dynamic - motor Coulomb friction [Tc⁺, Tc⁻]
    :type Tc: ndarray(2,)
    :param G: dynamic - gear ratio
    :type G: float

    A subclass of the DHLink class for a prismatic joint that holds all
    information related to a robot link such as kinematics parameters,
    rigid-body inertial parameters, motor and transmission parameters.

    The link transform is

    :math:`\mathbf{T}_{tx}(a_{i-1}) \cdot \mathbf{T}_{rx}(\alpha_{i-1}) \cdot \mathbf{T}_{rz}(\theta_i) \cdot \underbrace{\mathbf{T}_{tz}(q_i)}_{\mbox{variable}}`

    where :math:`q_i` is the joint variable.

    :references:
        - Robotics, Vision & Control, P. Corke, Springer 2011, Chap 7.

    :seealso: :func:`RevoluteMDH`, :func:`DHLink`, :func:`PrismaticDH`
    """  # noqa

    def __init__(
            self,
            theta=0.0,
            a=0.0,
            alpha=0.0,
            offset=0.0,
            qlim=None,
            flip=False,
            **kwargs
            ):

        d = 0.0
        sigma = 1
        mdh = True

        super().__init__(
            theta=theta, d=d, a=a, alpha=alpha, sigma=sigma, mdh=mdh,
            offset=offset, qlim=qlim, flip=flip, **kwargs)



from roboticstoolbox.models.URDF.Panda import Panda
from roboticstoolbox.models.URDF.Frankie import Frankie
from roboticstoolbox.models.URDF.UR3 import UR3
from roboticstoolbox.models.URDF.UR5 import UR5
from roboticstoolbox.models.URDF.UR10 import UR10
from roboticstoolbox.models.URDF.Puma560 import Puma560
from roboticstoolbox.models.URDF.px100 import px100
from roboticstoolbox.models.URDF.px150 import px150
from roboticstoolbox.models.URDF.rx150 import rx150
from roboticstoolbox.models.URDF.rx200 import rx200
from roboticstoolbox.models.URDF.vx300 import vx300
from roboticstoolbox.models.URDF.vx300s import vx300s
from roboticstoolbox.models.URDF.wx200 import wx200
from roboticstoolbox.models.URDF.wx250 import wx250
from roboticstoolbox.models.URDF.wx250s import wx250s
from roboticstoolbox.models.URDF.Mico import Mico
from roboticstoolbox.models.URDF.PR2 import PR2

__all__ = [
    'Panda',
    'Frankie',
    'UR3',
    'UR5',
    'UR10',
    'Puma560',
    'px100',
    'px150',
    'rx150',
    'rx200',
    'vx300',
    'vx300s',
    'wx200',
    'wx250',
    'wx250s',
    'Mico',
    'PR2'
]



from roboticstoolbox.models.ETS.Panda import Panda
from roboticstoolbox.models.ETS.Frankie import Frankie
from roboticstoolbox.models.ETS.Puma560 import Puma560
from roboticstoolbox.models.ETS.Planar_Y import Planar_Y
from roboticstoolbox.models.ETS.GenericSeven import GenericSeven


__all__ = [
    'Panda',
    'Frankie',
    'Puma560',
    'Planar_Y',
    'GenericSeven'
]



from roboticstoolbox import ETS
import re
import sympy
import numpy as np
from spatialmath import base

pi2 = base.pi() / 2
deg = base.pi() / sympy.Integer('180')

# PROGRESS
# subs2z does a bad thing in first phase, 2 subs it shouldnt make
class DHFactor(ETS):

    def __init__(self, axis=None, eta=None, **kwargs):

        super().__init__(axis=axis, eta=eta, **kwargs)

    @classmethod
    def parse(cls, s):
        et_re = re.compile(r"([RT][xyz])\(([^)]*)\)")

        # self.data = []

        jointnum = 0
        ets = DHFactor()

        for axis, eta in et_re.findall(s):
            if eta[0] == 'q':
                eta = None
                unit = None
                j = jointnum
                jointnum += 1
            else:
                # eta can be given as a variable or a number
                try:
                    # first attempt to create symbolic number
                    eta = sympy.Number(eta)
                except:
                    # failing that, a symbolic variable
                    eta = sympy.symbols(eta)
                if axis[0] == 'R':
                    # convert to degrees, assumed unit in input string
                    eta = eta * deg
                j = None

            if axis == 'Rx':
                e = DHFactor.rx(eta, j=j)
            elif axis == 'Ry':
                e = DHFactor.ry(eta, j=j)
            elif axis == 'Rz':
                e = DHFactor.rz(eta, j=j)
            elif axis == 'Tx':
                e = DHFactor.tx(eta, j=j)
            elif axis == 'Ty':
                e = DHFactor.ty(eta, j=j)
            elif axis == 'Tz':
                e = DHFactor.tz(eta, j=j)

            ets *= e
        
        return cls(ets)

    # ---------------------------------------------------------------------- #

    def simplify(self, mdh=False):

        self.merge()
        self.swap(mdh)
        self.merge()
        self.float_right()
        self.merge()
        # self.substitute_to_z()
        self.merge()
        print('inital merge and swap\n  ', self)
        print('main loop')
        for i in range(10):
            nchanges = 0

            nchanges += self.merge()
            nchanges += self.swap(mdh)
            nchanges += self.merge()
            print(ets)
            nchanges += self.eliminate_y()
            nchanges += self.merge()
            print(ets)

            if nchanges == 0:
                # try this then
                nchanges += self.substitute_to_z2()
            if nchanges == 0:
                break
        else:
            # too many iterations
            print('too many iterations')

    # ---------------------------------------------------------------------- #

    def float_right(self):
        """
        Attempt to 'float' translational terms as far to the right as
	 * possible and across joint boundaries.
        """
        
        nchanges = 0

        for i in range(len(self)):

            this = self[i]
            if this.isjoint or this.isrevolute:
                continue
            
            crossed = False
            for j in range(i+1, len(self)):
                next = self[j]

                if next.isprismatic:
                    continue
                if next.isrevolute and next.axis[1] == this.axis[1]:
                    crossed = True
                    continue
                break

            if crossed:
                del self[i]
                self.insert(j-1, this)
                nchanges += 1
                print('floated')

        return nchanges
    # ---------------------------------------------------------------------- #
    def swap(self, mdh=False):

        #  we want to sort terms into the order:
        # 	RZ
        # 	TX
        # 	TZ
        # 	RX

        def do_swap(this, next):
            if mdh:
                # modified DH
                return this.axis == 'Rx' and next.axis == 'tx' \
                    or this.axis == 'Ry' and next.axis == 'ty' \
                    or this.axis == 'Rz' and next.axis == 'tz' \
                    or this.axis == 'tz' and next.axis == 'tx'
            else:
                # standard DH

                # push constant translations through rotational joints
				# of the same type

                if this.axis == 'tz' and next.axis == 'tx':
                    return True

                if next.isjoint and (
                        this.axis == 'tx' and next.axis == 'Rx'
                        or this.axis == 'ty' and next.axis == 'Ry'
                        or this.axis == 'tz' and next.axis == 'Rz'):
                    return True
                elif not this.isjoint and (
                        this.axis == 'Rx' and next.axis == 'tx'
                        or this.axis == 'Ry' and next.axis == 'ty'):
                    return True

                if not this.isjoint and not next.isjoint and \
                        this.axis == 'tz' and next.axis == 'Rz':
                    return True

                #  move Ty terms to the right
                if this.axis == 'ty' and next.axis == 'tz' \
                        or this.axis == 'ty' and next.axis == 'tx':
                    return True

            return False

        total_changes = 0
        while True:
            nchanges = 0
            for i in range(len(self) - 1):
                this = self[i]
                next = self[i+1]

                if do_swap(this, next):
                    del self[i]
                    self.insert(i+1, this)
                    nchanges += 1
                    print(f"swapping {this} <--> {next}")
            if nchanges == 0:
                total_changes += nchanges
                break

        return total_changes

    # ---------------------------------------------------------------------- #


    def substitute_to_z(self):
        # substitute all non Z joint transforms according to rules
        nchanges = 0
        out = DHFactor()

        def subs_z(this):
            if this.axis == 'Rx':
                return DHFactor.ry(pi2) \
                     * DHFactor.rz(this.eta) \
                     * DHFactor.ry(-pi2)
            elif this.axis == 'Ry':
                return DHFactor.rx(-pi2) \
                    * DHFactor.rz(this.eta) \
                    * DHFactor.rx(pi2)
            elif this.axis == 'tx':
                return DHFactor.ry(pi2) \
                     * DHFactor.tz(this.eta) \
                     * DHFactor.ry(-pi2)
            elif this.axis == 'ty':
                return DHFactor.rx(-pi2) \
                     * DHFactor.tz(this.eta) \
                     * DHFactor.rx(pi2)

        for e in self:
            if not e.isjoint:
                out *= e
            else:
                # do a substitution
                new = DHFactor.subs_z(e)
                if new is None:
                    out *= e
                    continue
                else:
                    out *= new
                    nchanges += 1
                    print(f"subs2z: {e} := {new}")

        self.data = out.data
        return nchanges

    def substitute_to_z2(self):
        # substitute all non Z joint transforms according to rules
        nchanges = 0
        out = DHFactor()
        jointyet = False

        def subs_z(prev, this):
            if this.axis == 'Ry':
                return DHFactor.rz(pi2) \
                     * DHFactor.rx(this.eta) \
                     * DHFactor.rz(-pi2)
            elif this.axis == 'ty':
                if prev.axis == 'Rz':
                    return DHFactor.rz(pi2) \
                         * DHFactor.ty(this.eta) \
                         * DHFactor.r(-pi2)
                else:
                    return DHFactor.rx(-pi2) \
                         * DHFactor.ty(this.eta) \
                         * DHFactor.rx(pi2)

        for i in range(len(self)):
            this = self[i]
            if this.isjoint:
                jointyet = True
                continue
            
            if i == 0 or not jointyet:
                continue

            prev = self[i-1]

            new = subs_z(prev, this)

            if new is None:
                out *= this
                continue
            else:
                out *= new
                nchanges += 1
                print(f"subs2z2: {this} := {new}")

        self.data = out.data
        return nchanges
    # ---------------------------------------------------------------------- #

    def merge(self):

        def can_merge(prev, this):
            return prev.axis == this.axis and not (prev.isjoint and this.isjoint)

        out = DHFactor()
        nchanges = 0
        while len(self.data) > 0:
            this = self.pop(0)

            if len(self.data) > 0:
                next = self[0]

                if can_merge(this, next):
                    new = DHFactor.add(this, next)
                    out *= new
                    self.pop(0)  # remove next from the queue
                    print(f"Merging {this * next} to {new}")
                else:
                    out *= this
            else:
                out *= this

        self.data = out.data
        return nchanges
        # remove zeros

    @staticmethod
    def add(this, that):
        if this.isjoint and not that.isjoint:
            out = DHFactor(this)
            if out.eta is None:
                out.eta = that.eta
            else:
                out.eta += that.eta
        elif not this.isjoint and that.isjoint:
            out = DHFactor(that)
            if out.eta is None:
                out.eta = this.eta
            else:
                out.eta += this.eta
        elif not this.isjoint and not that.isjoint:
            out = DHFactor(this)
            if out.eta is None:
                out.eta = that.eta
            else:
                out.eta += that.eta
        else:
            raise ValueError('both ET cannot be joint variables')
        return out

    # ---------------------------------------------------------------------- #

    def eliminate_y(self):

        nchanges = 0
        out = DHFactor()
        jointyet = False

        def subs_y(prev, this):
            if this.isjoint or prev.isjoint:
                return None

            new = None
            if prev.axis == 'Rx' and this.axis == 'ty':  # RX.TY -> TZ.RX
                new = DHFactor.tx(prev.eta) * prev
            elif prev.axis == 'Rx' and this.axis == 'tz':  # RX.TZ -> TY.RX
                new = DHFactor.ty(-prev.eta) * prev
            elif prev.axis == 'Ry' and this.axis == 'tz':  # RY.TX-> TZ.RY
                new = DHFactor.tz(-prev.eta) * prev
            elif prev.axis == 'Ry' and this.axis == 'tz':  # RY.TZ-> TX.RY
                new = DHFactor.tx(prev.eta) * prev

            elif prev.axis == 'ty' and this.axis == 'Rx':  # TY.RX -> RX.TZ
                new = this * DHFactor.tz(-this.eta)
            elif prev.axis == 'tx' and this.axis == 'Rz':  # TX.RZ -> RZ.TY
                new = this * DHFactor.tz(this.eta)
            elif prev.axis == 'Ry' and this.axis == 'Rx':  # RY(Q).RX -> RX.RZ(-Q)
                new = this * DHFactor.Rz(-prev.eta)
            elif prev.axis == 'Rx' and this.axis == 'Ry':  # RX.RY -> RZ.RX
                new = DHFactor.Rz(this.eta) * prev
            elif prev.axis == 'Rz' and this.axis == 'Rx':  # RZ.RX -> RX.RY
                new = this * DHFactor.Ry(this.eta)
            return new

        for i in range(len(self)):
            this = self[i]
            jointyet = this.isjoint
            if i == 0 or not jointyet:  # leave initial const xform
                continue

            prev = self[i-1]

            new = subs_y(prev, this)
            if new is not None:
                self[i-1:i+1] = new
                nchanges += 1
                print(f"eliminate Y: {this} := {new}")


        return nchanges

    # ---------------------------------------------------------------------- #

    def __str__(self, q=None):
        """
        Pretty prints the ETS

        :param q: control how joint variables are displayed
        :type q: str
        :return: Pretty printed ETS
        :rtype: str

        ``q`` controls how the joint variables are displayed:

        - None, format depends on number of joint variables
            - one, display joint variable as q
            - more, display joint variables as q0, q1, ...
            - if a joint index was provided, use this value
        - "", display all joint variables as empty parentheses ``()``
        - "θ", display all joint variables as ``(θ)``
        - format string with passed joint variables ``(j, j+1)``, so "θ{0}"
          would display joint variables as θ0, θ1, ... while "θ{1}" would
          display joint variables as θ1, θ2, ...  ``j`` is either the joint
          index, if provided, otherwise a sequential value.

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> e = ETS.rz() * ETS.tx(1) * ETS.rz()
            >>> print(e[:2])
            >>> print(e)
            >>> print(e.__str__(""))
            >>> print(e.__str__("θ{0}"))  # numbering from 0
            >>> print(e.__str__("θ{1}"))  # numbering from 1
            >>> # explicit joint indices
            >>> e = ETS.rz(j=3) * ETS.tx(1) * ETS.rz(j=4)
            >>> print(e)
            >>> print(e.__str__("θ{0}"))

        .. note:: Angular parameters are converted to degrees, except if they
            are symbolic.

        Example:

        .. runblock:: pycon

            >>> from roboticstoolbox import ETS
            >>> from spatialmath.base import symbol
            >>> theta, d = symbol('theta, d')
            >>> e = ETS.rx(theta) * ETS.tx(2) * ETS.rx(45, 'deg') * \
            >>>     ETS.ry(0.2) * ETS.ty(d)
            >>> str(e)

        :SymPy: supported
        """

        # override this class from ETS
        es = []
        j = 0
        c = 0

        if q is None:
            if len(self.joints()) > 1:
                q = "q{0}"
            else:
                q = "q"
        q = "q{0}"
        # For et in the object, display it, data comes from properties
        # which come from the named tuple
        for et in self:

            s = f"{et.axis}("
            if et.isjoint:
                if q is not None:
                    if et.jindex is None:
                        _j = j
                    else:
                        _j = et.jindex
                    qvar = q.format(_j, _j+1) # lgtm [py/str-format/surplus-argument]  # noqa
                else:
                    qvar = ""
                if et.isflip:
                    s += f"-{qvar}"
                else:
                    s += f"{qvar}"
                j += 1

            if et.eta is not None:
                if et.isrevolute:
                    if base.issymbol(et.eta):
                        if s[-1] == "(":
                            s += f"{et.eta}"
                        else:
                            # adding to a previous value
                            if str(et.eta).startswith('-'):
                                s += f"{et.eta}"
                            else:
                                s += f"+{et.eta}"
                    else:
                        s += f"{et.eta * 180 / np.pi:.4g}°"

                elif et.isprismatic:
                    s += f"{et.eta}"

                elif et.isconstant:
                    s += f"C{c}"
                    c += 1
            s += ")"

            es.append(s)

        return " * ".join(es)
            
if __name__ == "__main__":  # pragram: no cover
    s = 'Tz(L1) Rz(q1) Ry(q2) Ty(L2) Tz(L3) Ry(q3) Tx(L4) Ty(L5) Tz(L6) Rz(q4) Ry(q5) Rz(q6)'

    ets = DHFactor.parse(s)
    print(ets)
    # ets.substitute_to_z()
    # print(ets)
    # ets.merge()
    # print(ets)

    ets.swap()
    ets.simplify()
    print(ets)
    # ets.merge()
    print(ets)




"""Utilities for URDF parsing.
"""
import numpy as np
import spatialmath as sm


def parse_origin(node):
    """Find the ``origin`` subelement of an XML node and convert it
    into a 4x4 homogenous transformation matrix.
    Parameters
    ----------
    node : :class`lxml.etree.Element`
        An XML node which (optionally) has a child node with the ``origin``
        tag.
    Returns
    -------
    matrix : (4,4) float
        The 4x4 homogneous transform matrix that corresponds to this node's
        ``origin`` child. Defaults to the identity matrix if no ``origin``
        child was found.
    """
    matrix = sm.SE3()
    rpy = np.array([0, 0, 0])
    origin_node = node.find('origin')
    if origin_node is not None:
        if 'xyz' in origin_node.attrib:
            t = np.fromstring(origin_node.attrib['xyz'], sep=' ')
            matrix = sm.SE3(t)
        if 'rpy' in origin_node.attrib:
            rpy = np.fromstring(origin_node.attrib['rpy'], sep=' ')
            matrix.A[:3, :3] = sm.SE3.RPY(rpy).R

    return matrix.A, rpy


def configure_origin(value):
    """Convert a value into a 4x4 transform matrix.
    Parameters
    ----------
    value : None, (6,) float, or (4,4) float
        The value to turn into the matrix.
        If (6,), interpreted as xyzrpy coordinates.
    Returns
    -------
    matrix : (4,4) float or None
        The created matrix.
    """
    if value is None:
        value = np.eye(4, dtype=np.float64)
    return value
